/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
static char sccsid[] = "@(#)display.c	5.1 (Berkeley) 6/6/85";
#endif not lint

/*
 * The window 'manager', initializes curses and handles the actual
 * displaying of text
 */

#include <stdio.h>
#include "talk.h"

/*
 * This local version contains some extra control codes:
 *	^N	Display menu
 *	^F	Clear window
 *	^B	Toggle beep
 *	^A	display <go ahead>\n
 *	^E	display <sorry to interrupt you>\n
 *	^X	display <thats ok>\n
 *	^D	display <over and out>\n
 */

static char *cc_goahead = "<go ahead>";
static char *cc_beepon  = "<beep on>";
static char *cc_beepoff = "<beep off>";
static char *cc_sorry   = "<sorry to interrupt you>";
static char *cc_thatsok = "<thats okay>";
static char *cc_overout	= "<over and out>";

xwin_t	my_win;
xwin_t	his_win;
WINDOW	*line_win;

int	curses_initialized = 0;
int	beep = 0;

/*
 * max HAS to be a function, it is called with
 * a argument of the form --foo at least once.
 */
max(a,b)
	int a, b;
{

	return (a > b ? a : b);
}

/*
 * Display some text on somebody's window, processing some control
 * characters while we are at it.
 */
talk_display(win, text, size)
	register xwin_t *win;
	register char *text;
	int size;
{
	register int i;
	char cch;

	for (i = 0; i < size; i++, text++) {
		if (*text == '\n') {
			xscroll(win, 0);
			continue;
		}
		/* erase character */
		if (*text == win->cerase) {
			wmove(win->x_win, win->x_line, max(--win->x_col, 0));
			getyx(win->x_win, win->x_line, win->x_col);
			waddch(win->x_win, ' ');
			wmove(win->x_win, win->x_line, win->x_col);
			getyx(win->x_win, win->x_line, win->x_col);
			continue;
		}
		/*
		 * On word erase search backwards until we find
		 * the beginning of a word or the beginning of
		 * the line.
		 */
		if (*text == win->werase) {
			int endcol, xcol, i, c;

			endcol = win->x_col;
			xcol = endcol - 1;
			while (xcol >= 0) {
				c = readwin(win->x_win, win->x_line, xcol);
				if (c != ' ')
					break;
				xcol--;
			}
			while (xcol >= 0) {
				c = readwin(win->x_win, win->x_line, xcol);
				if (c == ' ')
					break;
				xcol--;
			}
			wmove(win->x_win, win->x_line, xcol + 1);
			for (i = xcol + 1; i < endcol; i++)
				waddch(win->x_win, ' ');
			wmove(win->x_win, win->x_line, xcol + 1);
			getyx(win->x_win, win->x_line, win->x_col);
			continue;
		}
		/* line kill */
		if (*text == win->kill) {
			wmove(win->x_win, win->x_line, 0);
			wclrtoeol(win->x_win);
			getyx(win->x_win, win->x_line, win->x_col);
			continue;
		}
		/* clear screen */
		if (*text == '\f') {
			if (win == &my_win)
				wrefresh(curscr);
			continue;
		}
		if (win->x_col == COLS-1) {
			/* check for wraparound */
			xscroll(win, 0);
		}

		switch(*text) {		/* Expand special control codes */
			case '\001':
				if (win->x_col) xscroll(win,0);
				waddstr(win->x_win,cc_goahead);
				if (beep) putchar('\007');
				xscroll(win,0);
				continue;
			case '\002':
				beep = !beep;
				if (win->x_col) xscroll(win,0);
				waddstr(win->x_win,beep?cc_beepon:cc_beepoff);
				xscroll(win,0);
				continue;
			case '\004':
				if (win->x_col) xscroll(win,0);
				waddstr(win->x_win,cc_overout);
				xscroll(win,0);
				continue;
			case '\005':
				if (win->x_col) xscroll(win,0);
				waddstr(win->x_win,cc_sorry);
				xscroll(win,0);
				continue;
			case '\006':
				wclear(win->x_win);
				xscroll(win,-1);
				continue;
			case '\030':
				if (win->x_col) xscroll(win,0);
				waddstr(win->x_win,cc_thatsok);
				xscroll(win,0);
				continue;
			case '\016':
				wclear(win->x_win);
				xscroll(win,-1);
				waddstr(win->x_win,"\n^A\t\"<go ahead>\"");
				xscroll(win,0);
				waddstr(win->x_win,"^D\t\"<over and out>\"");
				xscroll(win,0);
				waddstr(win->x_win,"^E\t\"<sorry to interrupt you>\"");
				xscroll(win,0);
				waddstr(win->x_win,"^X\t\"<thats ok>\"");
				xscroll(win,0);
				waddstr(win->x_win,"^F\tClear your window");
				xscroll(win,0);
				waddstr(win->x_win,"^B\tToggle beep");
				xscroll(win,0);
				waddstr(win->x_win,"^C\tQuit talk");
				xscroll(win,0);
				waddstr(win->x_win,"^N\tThis menu");
				xscroll(win,0);
				xscroll(win,0); 
				continue;
		};

		if (*text == '\007' && beep)	/* just send out the beep */
			putchar(*text);
		else if (*text < ' ' && *text != '\t') {
			waddch(win->x_win, '^');
			getyx(win->x_win, win->x_line, win->x_col);
			if (win->x_col == COLS-1) /* check for wraparound */
				xscroll(win, 0);
			cch = (*text & 63) + 64;
			waddch(win->x_win, cch);
		} else
			waddch(win->x_win, *text);
		getyx(win->x_win, win->x_line, win->x_col);
	}
	wrefresh(win->x_win);
}

/*
 * Read the character at the indicated position in win
 */
readwin(win, line, col)
	WINDOW *win;
{
	int oldline, oldcol;
	register int c;

	getyx(win, oldline, oldcol);
	wmove(win, line, col);
	c = winch(win);
	wmove(win, oldline, oldcol);
	return (c);
}

/*
 * Scroll a window, blanking out the line following the current line
 * so that the current position is obvious
 */
xscroll(win, flag)
	register xwin_t *win;
	int flag;
{
	if (flag == -1) {
		wmove(win->x_win, 0, 0);
		win->x_line = 0;
		win->x_col = 0;
		return;
	}
	win->x_line = (win->x_line + 1) % win->x_nlines;
	win->x_col = 0;
	wmove(win->x_win, win->x_line, win->x_col);
	wclrtoeol(win->x_win);
	wmove(win->x_win, (win->x_line + 1) % win->x_nlines, win->x_col);
	wclrtoeol(win->x_win);
	wmove(win->x_win, win->x_line, win->x_col);
}

chat_display(win, text, size)	/* Display chars using the 'chat' interface */
register xwin_t *win;
register char *text;
int size;
{
    register int i, j, c;

    for (i=0; i<size; i++, text++) {

	c = *text;

	if (c == win->cerase && win->column) {	/* Character erase */
		printf("\b \b");
		win->column--;
	} else if (c == win->werase) {		/* Word erase */
		while (win->column && win->chatline[win->column-1]==' ') {
			putchar('\b');
			win->column--;
		}
		while (win->column && win->chatline[win->column-1]!=' ') {
			printf("\b \b");
			win->column--;
		}
	} else if (c == win->kill) {		/* Line erase */
		while (win->column) {
			printf("\b \b");
			win->column--;
		}
	} else if (c == win->redraw) {		/* Line redraw */
		if (c < ' ') {
			putchar('^');
			putchar(c+64);
		} else if (c < 0177)
			putchar(c);
		printf("\r\n");
		for(j=0; j<win->column; j++)
			putchar(win->chatline[j]);
	} else if (c == '\t') {			/* Tab display and expansion */
		putchar('\t');
		for (j=0; j<8-win->column%8 && win->column<CHAT_SIZ; j++)
			win->chatline[win->column+j] = ' ';
		win->column += j;
	} else					
		switch(c) {		/* Expand special control codes */
			case '\001':
				printf("%s%s%c\r\n", win->column? "\r\n": "",
					cc_goahead, beep? '\007': '\0');
				win->column = 0;
				break;
			case '\002':
				beep = !beep;
				printf("%s%s\r\n", win->column? "\r\n": "",
					beep? cc_beepon: cc_beepoff);
				break;
			case '\030':
				printf("%s%s\r\n", win->column? "\r\n": "",
					cc_thatsok);
				win->column = 0;
				break;
			case '\004':
				printf("%s%s\r\n", win->column? "\r\n": "",
					cc_overout);
				win->column = 0;
				break;
			case '\005':
				printf("%s%s\r\n", win->column? "\r\n": "",
					cc_sorry);
				win->column = 0;
				break;
			case '\006':
				win->column = 0;
				break;
			case '\016':
				if (win->column)
					printf("\r\n");
				printf("\r\n^A\t\"<go ahead>\"\r\n");
				printf("^D\t\"<over and out>\"\r\n");
				printf("^E\t\"<sorry to interrupt you>\"\r\n");
				printf("^X\t\"<thats okay>\"\r\n");
				printf("^F\tClear your window\r\n");
				printf("^B\tToggle beep\r\n");
				printf("^C\tQuit chat\r\n");
				printf("^N\tThis menu\r\n\r\n");
				break;
			case '\n':
				printf("\r\n");
				win->column = 0;
				break;
			case '\007':
				if (beep) {
					putchar(c);
					break;
				}
			default:	/* Just a character -- display it */
				/* Display control chars like "cat -v" does */

				if (c > 0177) {		/* A meta character! */
					sendchar('M', win);
					sendchar('-', win);
					c &= 0177;	/* zero the high bit */
				}
				if (c < ' ' || c == 0177) {	/* ctrl char */
					if (c=='\007' && beep)	/* a beep */
						putchar(c);
					else {		/* Display '^' first */
						sendchar('^', win);
						if (c == 0177)
							c = '?';
						else
							c += 64;
					}
				}
				sendchar(c, win);	/* Display character */
		}

		(void) fflush(stdout);	/* One character at-a-time display */
    }
}

sendchar(c, win)	/* Display char to tty and add it to users' line */
char c;
register xwin_t *win;
{
	putchar(c);

	if (win->column < CHAT_SIZ-1)	/* save char in users' line buffer */
		win->chatline[win->column++] = c;
}
