/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
static char sccsid[] = "@(#)io.c	5.1 (Berkeley) 6/6/85";
#endif not lint

/*
 * This file contains the I/O handling and the exchange of 
 * edit characters. This connection itself is established in
 * ctl.c
 */

#include "talk.h"
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>

#define A_LONG_TIME 10000000
#define STDIN_MASK (1<<fileno(stdin))	/* the bit mask for standard
					   input */
extern int errno;
/*
 * The routine to do the actual talking
 */
talk()
{
	register int read_template, sockt_mask, i, j;
	int read_set, nb;
	char buf[BUFSIZ];
	struct timeval wait;
	static char *chat_mesg = "[I am using the 'chat' interface]\n";

	if (my_interfac != I_CHAT)
		*chat_mesg = '\0';

	message("Connection established\007\007\007");
	current_line = 0;
	sockt_mask = (1<<sockt);

	/*
	 * Wait on both the other process (sockt_mask) and 
	 * standard input ( STDIN_MASK )
	 */
	read_template = sockt_mask | STDIN_MASK;
	forever {
		read_set = read_template;
		wait.tv_sec = A_LONG_TIME;
		wait.tv_usec = 0;
		nb = select(32, &read_set, 0, 0, &wait);
		if (nb <= 0) {
			if (errno == EINTR) {
				read_set = read_template;
				continue;
			}
			/* panic, we don't know what happened */
			p_error("Unexpected error from select");
			quit();
		}
		if (read_set & sockt_mask) { 
			/* There is data on sockt */
			nb = read(sockt, buf, sizeof buf);
			if (nb <= 0) {
				message("Connection closed. Exiting");
				quit();
			} else {	/* Don't accept ^B and ^N */
				for (i=j=0; i<nb; i++) {
					if (buf[i]=='\002' || buf[i]=='\016')
						continue;
					buf[j++] = buf[i];
				}
				if (j) {	/* use the correct interface */
					if (my_interfac == I_TALK)
						talk_display(&his_win, buf, j);
					else if (my_interfac == I_CHAT)
						chat_display(&his_win, buf, j);
				}
			}
		}
		if (read_set & STDIN_MASK) {
			/*
			 * We can't make the tty non_blocking, because
			 * curses's output routines would screw up
			 */
			ioctl(0, FIONREAD, (struct sgttyb *) &nb);
			nb = read(0, buf, nb);

			if (my_interfac == I_TALK)
				talk_display(&my_win, buf, nb);
			else if (my_interfac == I_CHAT) {
				for (i=0; i<nb; i++)	/* strip high bit */
					buf[i] &= 0177;
				chat_display(&my_win, buf, nb);

				/* Do not send the redraw, it looks stupid */
				for (i=j=0; i<nb; i++) {
					if (buf[i] != my_win.redraw)
						buf[j++] = buf[i];
				}
				nb = j;
			}

			if (*chat_mesg) {   /* inform them we are using chat */
				write(sockt, chat_mesg, strlen(chat_mesg));
				*chat_mesg = '\0';
			}

			/* might lose data here because sockt is non-blocking */
			write(sockt, buf, nb);
		}
	}
}

extern	int errno;
extern	int sys_nerr;
extern	char *sys_errlist[];

/*
 * p_error prints the system error message on the standard location
 * on the screen and then exits. (i.e. a curses version of perror)
 */
p_error(string) 
	char *string;
{
	char *sys;

	sys = "Unknown error";
	if (errno < sys_nerr)
		sys = sys_errlist[errno];

	if (my_interfac == I_TALK) {
		wmove(my_win.x_win, current_line%my_win.x_nlines, 0);
		wprintw(my_win.x_win, "[%s : %s (%d)]\n", string, sys, errno);
		wrefresh(my_win.x_win);
		move(LINES-1, 0);
		refresh();
		quit();
	} else if (my_interfac == I_CHAT)
		printf("[%s : %s (%d)]\r\n", string, sys, errno);
}

/*
 * Display string in the standard location
 */
message(string)
	char *string;
{

	if (my_interfac == I_TALK) {
		wmove(my_win.x_win, current_line%my_win.x_nlines, 0);
		wprintw(my_win.x_win, "[%s]\n", string);
		wrefresh(my_win.x_win);
	} else if (my_interfac == I_CHAT)
		printf("[%s]\r\n", string);
}
