/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
static char sccsid[] = "@(#)announce.c	5.3 (Berkeley) 3/13/86";
#endif not lint

#include <sys/types.h>
#include <sys/stat.h>
#include <sgtty.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <stdio.h>
#include <sys/wait.h>
#include <errno.h>
#include <syslog.h>
#include <strings.h>
#include <pwd.h>

#include <protocols/talkd.h>

#if PYRAMID
#define	ntohl(x)	(x)
#define	ntohs(x)	(x)
#define	htonl(x)	(x)
#define	htons(x)	(x)
#endif

extern	int errno;
extern	char hostname[];

/*
 * Announce an invitation to talk.
 *
 * Because the tty driver insists on attaching a terminal-less
 * process to any terminal that it writes on, we must fork a child
 * to protect ourselves
 */
announce(request, remote_machine, busy)
	CTL_MSG *request;
	char *remote_machine;
	CTL_BUSY *busy;
{
	int pid, val, status;
	int fd[2];

	if (pipe(fd) < 0)	/* need a pipe to return .busy from child */
		return(FAILED);

	if (pid = fork()) {
		/* we are the parent, so wait for the child */
		if (pid == -1) {	/* the fork failed */
			close(fd[0]);
			close(fd[1]);
			return (FAILED);
		}

		do {
			val = wait(&status);
			if (val == -1) {
				if (errno == EINTR)
					continue;
				/* shouldn't happen */
				close(fd[0]);
				close(fd[1]);
				syslog(LOG_WARNING, "announce: wait: %m");
				return (FAILED);
			}
		} while (val != pid);
		if (status&0377 > 0) {	/* we were killed by some signal */
			close(fd[0]);
			close(fd[1]);
			return (FAILED);
		}
		/* Get the second byte, this is the exit/return code */
		status = (status >> 8) & 0377;

		if (status == PERMISSION_DENIED)	/* get .busy file */
			get_busy(fd[0], busy);
		close(fd[0]);
		close(fd[1]);
		return(status);
	}
	/* we are the child, go and do it */
	_exit(announce_proc(request, remote_machine, fd[1]));
}

/*
 * See if the user is accepting messages. If so, announce that 
 * a talk is requested.
 */
announce_proc(request, remote_machine, fd_write)
	CTL_MSG *request;
	char *remote_machine;
	int fd_write;
{
	int pid, status;
	char full_tty[32];
	FILE *tf;
	struct stat stbuf;

	sprintf(full_tty, "/dev/%s", request->r_tty);
	if (access(full_tty, 0) != 0)
		return (FAILED);
	if ((tf = fopen(full_tty, "w")) == NULL) {
		refuse(request->r_name, fd_write);
		return (PERMISSION_DENIED);
	}
	/*
	 * On first tty open, the server will have
	 * it's pgrp set, so disconnect us from the
	 * tty before we catch a signal.
	 */
	ioctl(fileno(tf), TIOCNOTTY, (struct sgttyb *) 0);
	if (fstat(fileno(tf), &stbuf) < 0) {
		refuse(request->r_name, fd_write);
		return (PERMISSION_DENIED);
	}
	if ((stbuf.st_mode&020) == 0) {
		refuse(request->r_name, fd_write);
		return (PERMISSION_DENIED);
	}
	print_mesg(tf, request, remote_machine);
	fclose(tf);
	return (SUCCESS);
}

#define max(a,b) ( (a) > (b) ? (a) : (b) )
#define N_LINES 5
#define N_CHARS 120

/*
 * Build a block of characters containing the message. 
 * It is sent blank filled and in a single block to
 * try to keep the message in one piece if the recipient
 * in in vi at the time
 */
print_mesg(tf, request, remote_machine)
	FILE *tf;
	CTL_MSG *request;
	char *remote_machine;
{
	struct timeval clock;
	struct timezone zone;
	struct tm *localtime();
	struct tm *localclock;
	char line_buf[N_LINES][N_CHARS];
	int sizes[N_LINES];
	char big_buf[N_LINES*N_CHARS];
	char *bptr, *lptr;
	int i, j, max_size;

	i = 0;
	max_size = 0;
	gettimeofday(&clock, &zone);
	localclock = localtime( &clock.tv_sec );
	sprintf(line_buf[i], " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	sprintf(line_buf[i], "Message from NTalk_Daemon@%s at %d:%02d ...",
	hostname, localclock->tm_hour , localclock->tm_min );
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	sprintf(line_buf[i], "ntalk: connection requested by %s@%s.",
		request->l_name, remote_machine);
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	sprintf(line_buf[i], "ntalk: respond with:  ntalk %s@%s",
		request->l_name, remote_machine);
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	sprintf(line_buf[i], " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	bptr = big_buf;
	*bptr++ = ''; /* send something to wake them up */
	*bptr++ = '\r';	/* add a \r in case of raw mode */
	*bptr++ = '\n';
	for (i = 0; i < N_LINES; i++) {
		/* copy the line into the big buffer */
		lptr = line_buf[i];
		while (*lptr != '\0')
			*(bptr++) = *(lptr++);
		/* pad out the rest of the lines with blanks */
		for (j = sizes[i]; j < max_size + 2; j++) {
			*(bptr++) = ' ';
		}
		*(bptr++) = '\r';	/* add a \r in case of raw mode */
		*(bptr++) = '\n';
	}
	*bptr = '\0';
	fprintf(tf, big_buf);
	fflush(tf);
	ioctl(fileno(tf), TIOCNOTTY, (struct sgttyb *) 0);
}

refuse(r_name, fd_write)	/* try to find out why user is unsociable */
char *r_name;
int fd_write;
{
	register FILE *fpr;
	char bpth[100];			/* homedirectory/BUSYFLNM */
	char bufr[BUSYSIZE];		/* contents of BUSYFLNM */
	register int c, i=0;
	struct passwd *my_getpwnam();

	*bufr = '\0';
	(void) strcat(strcat(strcpy(bpth, my_getpwnam(r_name)->pw_dir), "/"),
		      BUSYFLNM);
	if ((fpr=fopen(bpth, "r")) != NULL) {	/* file exists */
		while ((c=getc(fpr)) != EOF && i<BUSYSIZE-1)
			bufr[i++] = c;		/* read file to buffer */
		bufr[i] = '\0';
		(void) fclose(fpr);
	}
	(void) write(fd_write, bufr, BUSYSIZE);	/* write to pipe */
}

get_busy(fd_read, busy)		/* read from pipe into CTL_BUSY */
int fd_read;
CTL_BUSY *busy;
{
	register int i;

	busy->chksum = (long) 0;	/* We also want a check sum */

	(void) read(fd_read, busy->busy, BUSYSIZE);	/* Read from pipe */

	for (i=0; i<BUSYSIZE && busy->busy[i]; i++)	/* Compute checksum */
		busy->chksum += (long) busy->busy[i];
	busy->chksum = htonl(busy->chksum);
}

/*
 *	I had to write my own 'getpwnam' because the one that is
 *	provided seems to have an obscure bug in it.  The problem
 *	was tracked down to 'endpwent' which calls 'dbm_close'
 *	which makes a call to 'free' (and promptly dies).  My
 *	solution was to simply eliminate the call to 'endpwent'
 *	because if this code was executed, the user must have had
 *	his messages off and the daemon will die (thus closing the
 *	database files).  Even if the daemon doesn't die quickly
 *	and this procedure is called again, 'setpwent' knows that
 *	the database is still open and just rewinds it.
 *
 *	Incidentally, when I used 'endpwent', talk responded:
 *
 *		[No connection yet]
 *		[Target machine is too confused to talk to us]
 *		<exit>
 */

struct passwd *my_getpwnam(usrnam)	/* getpwnam without call to endpwent */
char *usrnam;
{
	struct passwd *pw;

	(void) setpwent();		/* Open password file */

	while ((pw = getpwent()) && strcmp(usrnam, pw->pw_name))
		;	/* Find the right entry */

	return(pw);	/* It MUST exist (user is logged in), just return it */
}
