/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1983 Regents of the University of California.\n\
 All rights reserved.\n";
#endif not lint

#ifndef lint
static char sccsid[] = "@(#)talkd.c	5.2 (Berkeley) 3/13/86";
#endif not lint

/*
 * The top level of the daemon, the format is heavily borrowed
 * from rwhod.c. Basically: find out who and where you are; 
 * disconnect all descriptors and ttys, and then endless
 * loop on waiting for and processing requests
 */
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>

#include <protocols/talkd.h>

#ifdef NO_INETD
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <netdb.h>
#endif

CTL_MSG		request;
CTL_RESPONSE	response;
CTL_BUSY	busy;

int	sockt = 0;
int	debug = 0;
int	timeout();
long	lastmsgtime;

char	hostname[32];

#ifdef NO_INETD
struct	sockaddr_in sin = { AF_INET };
u_short	port = 0;
#endif

#define TIMEOUT 30
#define MAXIDLE 120

main(argc, argv)
	int argc;
	char *argv[];
{
#ifdef NO_INETD
	struct servent *sp, *getservbyname();
	struct sockaddr_in from;
	int fromlen;
#endif
	register CTL_MSG *mp = &request;
	int cc;

	if (getuid()) {
		fprintf(stderr, "%s: getuid: not super-user", argv[0]);
		exit(1);
	}
#if (SUN3_1 | SUN3_2 | PYRAMID | ULTRIX1_2)
	openlog("talkd", LOG_PID);
#else
	openlog("talkd", LOG_PID, LOG_DAEMON);
#endif

	if (gethostname(hostname, sizeof (hostname) - 1) < 0) {
		syslog(LOG_ERR, "gethostname: %m");
		_exit(1);
	}
	if (chdir("/dev") < 0) {
		syslog(LOG_ERR, "chdir: /dev: %m");
		_exit(1);
	}
	if (argc > 1 && strcmp(argv[1], "-d") == 0)
		debug = 1;

#ifdef NO_INETD
	if ((sp=getservbyname("ntalk", "udp")) == (struct servent *) 0) {
		syslog(LOG_ERR, "gethostbyname: %m");
		_exit(1);
	}
	port = sp->s_port;

	if ((sockt=socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		syslog(LOG_ERR, "socket: %m");
		_exit(1);
	}

	bzero((char *)&sin, sizeof(sin));
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = port;
	sin.sin_family = AF_INET;

	if (bind(sockt, (caddr_t)&sin, sizeof(sin)) < 0) {
		syslog(LOG_ERR, "bind: %m");
		_exit(1);
	}

	if (fork())
		_exit(0);
	setup_descriptors(sockt);

#else
	signal(SIGALRM, timeout);
	alarm(TIMEOUT);
#endif
	for (;;) {
		extern int errno;

#ifdef NO_INETD
		fromlen = sizeof(from);
		cc = recvfrom(sockt, (char *)mp,
		    sizeof (*mp), 0, &from, &fromlen);
#else
		cc = recv(sockt, (char *)mp, sizeof (*mp), 0);
#endif
		if (cc != sizeof (*mp)) {
			if (cc < 0 && errno != EINTR)
				syslog(LOG_WARNING, "recv: %m");
			continue;
		}
		lastmsgtime = time(0);
		process_request(mp, &response, &busy);
		/* can block here, is this what I want? */
		cc = sendto(sockt, (char *)&response,
		    sizeof (response), 0, &mp->ctl_addr, sizeof (mp->ctl_addr));
		if (cc != sizeof (response))
			syslog(LOG_WARNING, "sendto: %m");

		if (response.answer == PERMISSION_DENIED) {
			/* send reason why this party is refusing messages */

			cc = sendto(sockt, (char *) &busy, sizeof(busy), 0,
				&request.ctl_addr, sizeof(request.ctl_addr));
			if (cc != sizeof(busy))
				syslog(LOG_WARNING, "sendto: %m");
		}
	}
}

timeout()
{

	if (time(0) - lastmsgtime >= MAXIDLE)
		_exit(0);
	alarm(TIMEOUT);
}

#ifdef NO_INETD
setup_descriptors(sockt)
int sockt;
{
	int i, tty;

	if (sockt != 0)
		(void) close(0);

	i = open("/dev/null", 0);
	if (sockt != 1)
		(void) dup2(i, 1);
	if (sockt != 2)
		(void) dup2(i, 2);

	if ((tty=open("/dev/tty", 0)) != -1) {
		ioctl(tty, TIOCNOTTY, (struct sgttyb *) 0);
		(void) close(tty);
	} else
		setpgrp(0, getpid());
}
#endif
